import gzip
import json
import struct
import sys
from collections.abc import Callable
from concurrent.futures import Executor
from concurrent.futures.thread import ThreadPoolExecutor
from dataclasses import dataclass
from pathlib import Path

from sonolus.backend.mode import Mode
from sonolus.build.compile import CompileCache, compile_mode
from sonolus.script.archetype import _BaseArchetype
from sonolus.script.bucket import Buckets
from sonolus.script.effect import Effects
from sonolus.script.engine import EngineData, empty_play_mode, empty_preview_mode, empty_tutorial_mode, empty_watch_mode
from sonolus.script.instruction import (
    TutorialInstructionIcons,
    TutorialInstructions,
)
from sonolus.script.internal.callbacks import (
    navigate_callback,
    preprocess_callback,
    update_callback,
    update_spawn_callback,
)
from sonolus.script.internal.context import ProjectContextState, ReadOnlyMemory
from sonolus.script.options import Options
from sonolus.script.particle import Particles
from sonolus.script.project import BuildConfig
from sonolus.script.sprite import Skin
from sonolus.script.ui import UiConfig

type JsonValue = bool | int | float | str | list[JsonValue] | dict[str, JsonValue] | None


@dataclass
class PackagedEngine:
    configuration: bytes
    play_data: bytes
    watch_data: bytes
    preview_data: bytes
    tutorial_data: bytes
    rom: bytes

    def write(self, path: Path):
        path.mkdir(parents=True, exist_ok=True)
        (path / "EngineConfiguration").write_bytes(self.configuration)
        (path / "EnginePlayData").write_bytes(self.play_data)
        (path / "EngineWatchData").write_bytes(self.watch_data)
        (path / "EnginePreviewData").write_bytes(self.preview_data)
        (path / "EngineTutorialData").write_bytes(self.tutorial_data)
        (path / "EngineRom").write_bytes(self.rom)

    def read(self, path: Path):
        self.configuration = (path / "EngineConfiguration").read_bytes()
        self.play_data = (path / "EnginePlayData").read_bytes()
        self.watch_data = (path / "EngineWatchData").read_bytes()
        self.preview_data = (path / "EnginePreviewData").read_bytes()
        self.tutorial_data = (path / "EngineTutorialData").read_bytes()
        self.rom = (path / "EngineRom").read_bytes()


def no_gil() -> bool:
    return sys.version_info >= (3, 13) and not sys._is_gil_enabled()


def package_engine(
    engine: EngineData,
    config: BuildConfig | None = None,
    cache: CompileCache | None = None,
    project_state: ProjectContextState | None = None,
):
    if cache is None:
        cache = CompileCache()

    config = config or BuildConfig()
    if project_state is None:
        project_state = ProjectContextState.from_build_config(config)
    configuration = build_engine_configuration(engine.options, engine.ui)
    if no_gil():
        # process_cpu_count is available in Python 3.13+
        from os import process_cpu_count

        # Need a worker for each mode (so +4) plus at least one to do work
        thread_pool = ThreadPoolExecutor(4 + max(1, min(8, (process_cpu_count() or 1))))
    else:
        thread_pool = None

    play_mode = engine.play if config.build_play else empty_play_mode()
    watch_mode = engine.watch if config.build_watch else empty_watch_mode()
    preview_mode = engine.preview if config.build_preview else empty_preview_mode()
    tutorial_mode = engine.tutorial if config.build_tutorial else empty_tutorial_mode()

    if thread_pool is not None:
        futures = {
            "play": thread_pool.submit(
                build_play_mode,
                archetypes=play_mode.archetypes,
                skin=play_mode.skin,
                effects=play_mode.effects,
                particles=play_mode.particles,
                buckets=play_mode.buckets,
                project_state=project_state,
                config=config,
                thread_pool=thread_pool,
                cache=cache,
            ),
            "watch": thread_pool.submit(
                build_watch_mode,
                archetypes=watch_mode.archetypes,
                skin=watch_mode.skin,
                effects=watch_mode.effects,
                particles=watch_mode.particles,
                buckets=watch_mode.buckets,
                project_state=project_state,
                update_spawn=watch_mode.update_spawn,
                config=config,
                thread_pool=thread_pool,
                cache=cache,
            ),
            "preview": thread_pool.submit(
                build_preview_mode,
                archetypes=preview_mode.archetypes,
                skin=preview_mode.skin,
                project_state=project_state,
                config=config,
                thread_pool=thread_pool,
                cache=cache,
            ),
            "tutorial": thread_pool.submit(
                build_tutorial_mode,
                skin=tutorial_mode.skin,
                effects=tutorial_mode.effects,
                particles=tutorial_mode.particles,
                instructions=tutorial_mode.instructions,
                instruction_icons=tutorial_mode.instruction_icons,
                preprocess=tutorial_mode.preprocess,
                navigate=tutorial_mode.navigate,
                update=tutorial_mode.update,
                project_state=project_state,
                config=config,
                thread_pool=thread_pool,
                cache=cache,
            ),
        }

        play_data = futures["play"].result()
        watch_data = futures["watch"].result()
        preview_data = futures["preview"].result()
        tutorial_data = futures["tutorial"].result()
    else:
        play_data = build_play_mode(
            archetypes=play_mode.archetypes,
            skin=play_mode.skin,
            effects=play_mode.effects,
            particles=play_mode.particles,
            buckets=play_mode.buckets,
            project_state=project_state,
            config=config,
            thread_pool=None,
            cache=cache,
        )
        watch_data = build_watch_mode(
            archetypes=watch_mode.archetypes,
            skin=watch_mode.skin,
            effects=watch_mode.effects,
            particles=watch_mode.particles,
            buckets=watch_mode.buckets,
            project_state=project_state,
            update_spawn=watch_mode.update_spawn,
            config=config,
            thread_pool=None,
            cache=cache,
        )
        preview_data = build_preview_mode(
            archetypes=preview_mode.archetypes,
            skin=preview_mode.skin,
            project_state=project_state,
            config=config,
            thread_pool=None,
            cache=cache,
        )
        tutorial_data = build_tutorial_mode(
            skin=tutorial_mode.skin,
            effects=tutorial_mode.effects,
            particles=tutorial_mode.particles,
            instructions=tutorial_mode.instructions,
            instruction_icons=tutorial_mode.instruction_icons,
            preprocess=tutorial_mode.preprocess,
            navigate=tutorial_mode.navigate,
            update=tutorial_mode.update,
            project_state=project_state,
            config=config,
            thread_pool=None,
            cache=cache,
        )

    return PackagedEngine(
        configuration=package_data(configuration),
        play_data=package_data(play_data),
        watch_data=package_data(watch_data),
        preview_data=package_data(preview_data),
        tutorial_data=package_data(tutorial_data),
        rom=package_rom(project_state.rom),
    )


def validate_engine(
    engine: EngineData,
    config: BuildConfig | None = None,
    project_state: ProjectContextState | None = None,
):
    config = config or BuildConfig()
    if project_state is None:
        project_state = ProjectContextState.from_build_config(config)

    play_mode = engine.play if config.build_play else empty_play_mode()
    watch_mode = engine.watch if config.build_watch else empty_watch_mode()
    preview_mode = engine.preview if config.build_preview else empty_preview_mode()
    tutorial_mode = engine.tutorial if config.build_tutorial else empty_tutorial_mode()

    build_play_mode(
        archetypes=play_mode.archetypes,
        skin=play_mode.skin,
        effects=play_mode.effects,
        particles=play_mode.particles,
        buckets=play_mode.buckets,
        project_state=project_state,
        config=config,
        thread_pool=None,
        validate_only=True,
    )
    build_watch_mode(
        archetypes=watch_mode.archetypes,
        skin=watch_mode.skin,
        effects=watch_mode.effects,
        particles=watch_mode.particles,
        buckets=watch_mode.buckets,
        project_state=project_state,
        update_spawn=watch_mode.update_spawn,
        config=config,
        thread_pool=None,
        validate_only=True,
    )
    build_preview_mode(
        archetypes=preview_mode.archetypes,
        skin=preview_mode.skin,
        project_state=project_state,
        config=config,
        thread_pool=None,
        validate_only=True,
    )
    build_tutorial_mode(
        skin=tutorial_mode.skin,
        effects=tutorial_mode.effects,
        particles=tutorial_mode.particles,
        instructions=tutorial_mode.instructions,
        instruction_icons=tutorial_mode.instruction_icons,
        preprocess=tutorial_mode.preprocess,
        navigate=tutorial_mode.navigate,
        update=tutorial_mode.update,
        project_state=project_state,
        config=config,
        thread_pool=None,
        validate_only=True,
    )


def build_engine_configuration(
    options: Options,
    ui: UiConfig,
) -> JsonValue:
    result = {
        "options": [option.to_dict() for option in options._options_],
        "ui": ui.to_dict(),
    }
    if replay_fallback_option_names := getattr(options, "replay_fallback_option_names", None):
        result["replayFallbackOptionNames"] = list(replay_fallback_option_names)
    return result


def build_play_mode(
    archetypes: list[type[_BaseArchetype]],
    skin: Skin,
    effects: Effects,
    particles: Particles,
    buckets: Buckets,
    project_state: ProjectContextState,
    config: BuildConfig,
    thread_pool: Executor | None = None,
    validate_only: bool = False,
    cache: CompileCache | None = None,
):
    return {
        **compile_mode(
            mode=Mode.PLAY,
            project_state=project_state,
            archetypes=archetypes,
            global_callbacks=None,
            passes=config.passes,
            thread_pool=thread_pool,
            validate_only=validate_only,
            cache=cache,
        ),
        "skin": build_skin(skin),
        "effect": build_effects(effects),
        "particle": build_particles(particles),
        "buckets": build_buckets(buckets),
    }


def build_watch_mode(
    archetypes: list[type[_BaseArchetype]],
    skin: Skin,
    effects: Effects,
    particles: Particles,
    buckets: Buckets,
    project_state: ProjectContextState,
    update_spawn: Callable[[], float],
    config: BuildConfig,
    thread_pool: Executor | None = None,
    validate_only: bool = False,
    cache: CompileCache | None = None,
):
    return {
        **compile_mode(
            mode=Mode.WATCH,
            project_state=project_state,
            archetypes=archetypes,
            global_callbacks=[(update_spawn_callback, update_spawn)],
            passes=config.passes,
            thread_pool=thread_pool,
            validate_only=validate_only,
            cache=cache,
        ),
        "skin": build_skin(skin),
        "effect": build_effects(effects),
        "particle": build_particles(particles),
        "buckets": build_buckets(buckets),
    }


def build_preview_mode(
    archetypes: list[type[_BaseArchetype]],
    skin: Skin,
    project_state: ProjectContextState,
    config: BuildConfig,
    thread_pool: Executor | None = None,
    validate_only: bool = False,
    cache: CompileCache | None = None,
):
    return {
        **compile_mode(
            mode=Mode.PREVIEW,
            project_state=project_state,
            archetypes=archetypes,
            global_callbacks=None,
            passes=config.passes,
            thread_pool=thread_pool,
            validate_only=validate_only,
            cache=cache,
        ),
        "skin": build_skin(skin),
    }


def build_tutorial_mode(
    skin: Skin,
    effects: Effects,
    particles: Particles,
    instructions: TutorialInstructions,
    instruction_icons: TutorialInstructionIcons,
    preprocess: Callable[[], None],
    navigate: Callable[[], None],
    update: Callable[[], None],
    project_state: ProjectContextState,
    config: BuildConfig,
    thread_pool: Executor | None = None,
    validate_only: bool = False,
    cache: CompileCache | None = None,
):
    return {
        **compile_mode(
            mode=Mode.TUTORIAL,
            project_state=project_state,
            archetypes=[],
            global_callbacks=[
                (preprocess_callback, preprocess),
                (navigate_callback, navigate),
                (update_callback, update),
            ],
            passes=config.passes,
            thread_pool=thread_pool,
            validate_only=validate_only,
            cache=cache,
        ),
        "skin": build_skin(skin),
        "effect": build_effects(effects),
        "particle": build_particles(particles),
        "instruction": build_instructions(instructions, instruction_icons),
    }


def build_skin(skin: Skin) -> JsonValue:
    return {
        "renderMode": skin.render_mode,
        "sprites": [{"name": name, "id": i} for i, name in enumerate(skin._sprites_)],
    }


def build_effects(effects: Effects) -> JsonValue:
    return {"clips": [{"name": name, "id": i} for i, name in enumerate(effects._effects_)]}


def build_particles(particles: Particles) -> JsonValue:
    return {"effects": [{"name": name, "id": i} for i, name in enumerate(particles._particles_)]}


def build_buckets(buckets: Buckets) -> JsonValue:
    return [bucket.to_dict() for bucket in buckets._buckets_]


def build_instructions(instructions: TutorialInstructions, instruction_icons: TutorialInstructionIcons) -> JsonValue:
    return {
        "texts": [{"name": name, "id": i} for i, name in enumerate(instructions._instructions_)],
        "icons": [{"name": name, "id": i} for i, name in enumerate(instruction_icons._instruction_icons_)],
    }


def package_rom(rom: ReadOnlyMemory) -> bytes:
    values = rom.values or [0]
    output = bytearray()

    for value in values:
        output.extend(struct.pack("<f", value))

    return gzip.compress(bytes(output), mtime=0)


def package_data(value: JsonValue) -> bytes:
    json_data = json.dumps(value, separators=(",", ":")).encode("utf-8")
    return gzip.compress(json_data, mtime=0)


def unpackage_data(data: bytes) -> JsonValue:
    json_data = gzip.decompress(data)
    return json.loads(json_data)


def print_table(rows: list[list[str]]) -> None:
    if not rows:
        return
    widths = [max(len(row[i]) for row in rows) for i in range(len(rows[0]))]
    for row in rows:
        print("  ".join(cell.ljust(widths[i]) for i, cell in enumerate(row)))


def print_visit_stats(stats):
    ordered = sorted(stats.items(), key=lambda item: item[1].own_time, reverse=True)
    rows = [
        [f"{name}:", f"{stat.own_time // 1_000_000}ms", f"{stat.total_time // 1_000_000}ms", f"{stat.call_count} calls"]
        for name, stat in ordered
    ]
    print_table(rows)



import random as pyrand
from collections.abc import MutableSequence, Sequence

from sonolus.backend.ops import Op
from sonolus.script.internal.native import native_function
from sonolus.script.values import copy


@native_function(Op.Random)
def _random_float(a: float, b: float) -> float:
    """Returns a random float between a and b, inclusive."""
    return pyrand.uniform(a, b)


@native_function(Op.RandomInteger)
def _random_integer(a: int, b: int) -> int:
    """Returns a random integer between a (inclusive) and b (exclusive)."""
    return pyrand.randrange(a, b)


def _randrange(start: int, stop: int | None = None, step: int = 1) -> int:
    if stop is None:
        stop = start
        start = 0
    range_len = (stop - start + step - 1) // step
    return start + step * _random_integer(0, range_len)


def _randint(a: int, b: int) -> int:
    return _random_integer(a, b + 1)


def _choice[T](seq: Sequence[T]) -> T:
    return seq[_randrange(len(seq))]


def _swap(seq: MutableSequence, i: int, j: int) -> None:
    temp = copy(seq[i])
    seq[i] = seq[j]
    seq[j] = temp


def _shuffle[T: MutableSequence](seq: T) -> None:
    i = len(seq) - 1
    while i > 0:
        j = _randrange(i + 1)
        _swap(seq, i, j)
        i -= 1


def _random() -> float:
    # The end needs to exclude 1, and Sonolus uses 32-bit floats.
    return _random_float(0.0, 0.99999994)


def _uniform(a: float, b: float) -> float:
    return _random_float(a, b)


RANDOM_BUILTIN_IMPLS = {
    id(pyrand.randrange): _randrange,
    id(pyrand.randint): _randint,
    id(pyrand.choice): _choice,
    id(pyrand.shuffle): _shuffle,
    id(pyrand.random): _random,
    id(pyrand.uniform): _uniform,
}



from math import pi

from hypothesis import assume, given
from hypothesis import strategies as st

from sonolus.script.vec import Vec2, angle_diff, signed_angle_diff
from tests.script.conftest import is_close, run_and_validate

floats = st.floats(min_value=-999, max_value=999, allow_nan=False, allow_infinity=False)
nonzero_floats = floats.filter(lambda x: abs(x) > 1e-2)
angles = st.floats(min_value=-pi, max_value=pi, allow_nan=False, allow_infinity=False)


def test_magnitude():
    def fn():
        v = Vec2(3, 4)
        return v.magnitude

    assert run_and_validate(fn) == 5


def test_angle():
    def fn():
        v = Vec2(1, 1)
        return v.angle

    assert is_close(run_and_validate(fn), 0.25 * pi)


def test_dot():
    def fn():
        v = Vec2(1, 2)
        u = Vec2(3, 4)
        return v.dot(u)

    assert run_and_validate(fn) == 11


def test_rotate():
    def fn():
        v = Vec2(1, 1)
        return v.rotate(0.25 * pi)

    result = run_and_validate(fn)
    assert is_close(result.x, 0)
    assert is_close(result.y, 2**0.5)


def test_rotate_about():
    def fn():
        v = Vec2(1, 1)
        pivot = Vec2(1, 0)
        return v.rotate_about(0.5 * pi, pivot)

    assert is_close(run_and_validate(fn).magnitude, 0)


def test_add():
    def fn():
        v = Vec2(1, 2)
        u = Vec2(3, 4)
        return v + u

    assert run_and_validate(fn) == Vec2(4, 6)


def test_sub():
    def fn():
        v = Vec2(1, 2)
        u = Vec2(3, 4)
        return v - u

    assert run_and_validate(fn) == Vec2(-2, -2)


def test_mul_vec():
    def fn():
        v = Vec2(1, 2)
        u = Vec2(3, 4)
        return v * u

    assert run_and_validate(fn) == Vec2(3, 8)


def test_mul_float():
    def fn():
        v = Vec2(1, 2)
        return v * 3

    assert run_and_validate(fn) == Vec2(3, 6)


def test_div_vec():
    def fn():
        v = Vec2(1, 2)
        u = Vec2(3, 4)
        return v / u

    assert run_and_validate(fn) == Vec2(1 / 3, 2 / 4)


def test_div_float():
    def fn():
        v = Vec2(1, 2)
        return v / 3

    assert run_and_validate(fn) == Vec2(1 / 3, 2 / 3)


def test_neg():
    def fn():
        v = Vec2(1, 2)
        return -v

    assert run_and_validate(fn) == Vec2(-1, -2)


def test_equal():
    def fn():
        v = Vec2(1, 2)
        u = Vec2(1, 2)
        return v == u

    assert run_and_validate(fn)


def test_not_equal():
    def fn():
        v = Vec2(1, 2)
        u = Vec2(3, 4)
        return v != u

    assert run_and_validate(fn)


@given(angles)
def test_unit_magnitude(angle):
    def fn():
        return Vec2.unit(angle).magnitude

    result = run_and_validate(fn)
    assert is_close(result, 1.0)


@given(angles)
def test_unit_angle(angle):
    def fn():
        return Vec2.unit(angle).angle

    result = run_and_validate(fn)
    assert is_close(result, angle)


@given(floats, floats)
def test_vec2_normalize_maintains_angle(x, y):
    assume(abs(x) > 1e-2 or abs(y) > 1e-2)

    def fn():
        v = Vec2(x, y)
        return angle_diff(v.angle, v.normalize().angle)

    diff = run_and_validate(fn)
    assert is_close(diff, 0)


@given(floats, floats)
def test_vec2_normalize_or_zero_maintains_angle(x, y):
    assume(abs(x) > 1e-2 or abs(y) > 1e-2)

    def fn():
        v = Vec2(x, y)
        return angle_diff(v.angle, v.normalize_or_zero().angle)

    diff = run_and_validate(fn)
    assert is_close(diff, 0)


def test_vec2_zero_normalize_or_zero():
    def fn():
        v = Vec2(0, 0)
        return v.normalize_or_zero()

    assert run_and_validate(fn) == Vec2(0, 0)


def test_angle_diff_examples():
    assert is_close(angle_diff(0, 0), 0)
    assert is_close(angle_diff(0, pi), pi)
    assert is_close(angle_diff(0, 2 * pi), 0)
    assert is_close(angle_diff(0, pi / 4), pi / 4)
    assert is_close(angle_diff(pi / 2, -pi / 2), pi)
    assert is_close(angle_diff(0, 3 * pi / 2), pi / 2)
    assert is_close(angle_diff(-pi / 4, pi / 4), pi / 2)


@given(floats, floats)
def test_angle_diff_in_range(a, b):
    result = angle_diff(a, b)
    assert 0 <= result <= pi


@given(floats)
def test_angle_diff_identity(angle):
    result = angle_diff(angle, angle)
    assert result == 0


@given(floats, floats)
def test_angle_diff_commutative(a, b):
    result1 = angle_diff(a, b)
    result2 = angle_diff(b, a)
    assert is_close(result1, result2)


@given(floats, st.floats(min_value=-pi, max_value=pi, allow_nan=False, allow_infinity=False))
def test_angle_diff_delta(angle, delta):
    result = angle_diff(angle, angle + delta)
    assert is_close(result, abs(delta))


def test_signed_angle_diff_examples():
    assert is_close(signed_angle_diff(0, 0), 0)
    assert is_close(signed_angle_diff(pi / 4, 0), pi / 4)
    assert is_close(signed_angle_diff(0, pi / 4), -pi / 4)
    assert is_close(signed_angle_diff(0, 2 * pi), 0)
    assert is_close(signed_angle_diff(pi / 2, 0), pi / 2)
    assert is_close(signed_angle_diff(0, 3 * pi / 4), -3 * pi / 4)
    assert is_close(signed_angle_diff(3 * pi / 4, -3 * pi / 4), -pi / 2)
    assert is_close(signed_angle_diff(pi, 0), -pi)


@given(floats, floats)
def test_signed_angle_diff_in_range(a, b):
    result = signed_angle_diff(a, b)
    assert -pi <= result <= pi


@given(floats)
def test_signed_angle_diff_identity(angle):
    result = signed_angle_diff(angle, angle)
    assert result == 0


@given(floats, floats)
def test_signed_angle_diff_anticommutative(a, b):
    result1 = signed_angle_diff(a, b)
    result2 = signed_angle_diff(b, a)
    assert is_close(result1, -result2)


@given(floats, st.floats(min_value=-pi, max_value=pi, exclude_max=True, allow_nan=False, allow_infinity=False))
def test_signed_angle_diff_delta(angle, delta):
    result = signed_angle_diff(angle + delta, angle)
    assert is_close(result, delta) or (is_close(abs(result), pi) and is_close(abs(delta), pi))



# Resolves a circular import in some cases.
# While restructuring imports could be a better solution, there are some performance implications on hot paths.
from sonolus.backend import visitor as visitor



# ruff: noqa: PT019
import random
from collections import defaultdict
from collections.abc import Callable

from hypothesis import given
from hypothesis import strategies as st

import sonolus.script.internal.random as srandom
from sonolus.script.array import Array
from sonolus.script.debug import assert_true
from sonolus.script.values import copy
from tests.script.conftest import run_compiled

# Strategies
ints = st.integers(min_value=-999, max_value=999)
floats = st.floats(min_value=-999, max_value=999, allow_infinity=False, allow_nan=False)
lists = st.lists(ints, min_size=1, max_size=20)


@given(st.random_module())
def test_randrange_basic(_r):
    def fn():
        value = random.randrange(10)
        assert_true(0 <= value < 10)
        return value

    result = run_compiled(fn)
    assert 0 <= result < 10


@given(st.random_module(), st.integers(min_value=-100, max_value=100), st.integers(min_value=1, max_value=100))
def test_randrange_with_start_stop(_r, start, width):
    stop = start + width

    def fn():
        value = random.randrange(start, stop)
        assert_true(start <= value < stop)
        return value

    result = run_compiled(fn)
    assert start <= result < stop


@given(
    st.random_module(),
    st.integers(min_value=-100, max_value=100),
    st.integers(min_value=1, max_value=100),
    st.integers(min_value=1, max_value=10),
)
def test_randrange_with_step(_r, start, width, step):
    stop = start + width

    def fn():
        value = random.randrange(start, stop, step)
        assert_true(start <= value < stop)
        assert_true((value - start) % step == 0)
        return value

    result = run_compiled(fn)
    assert start <= result < stop
    assert (result - start) % step == 0


@given(st.random_module(), st.integers(min_value=-100, max_value=100), st.integers(min_value=0, max_value=100))
def test_randint(_r, a, width):
    b = a + width

    def fn():
        value = random.randint(a, b)
        assert_true(a <= value <= b)
        return value

    result = run_compiled(fn)
    assert a <= result <= b


@given(st.random_module(), lists)
def test_choice(_r, values_list):
    values = Array(*values_list)

    def fn():
        value = random.choice(values)
        return value

    result = run_compiled(fn)
    assert result in values_list


@given(st.random_module(), lists)
def test_shuffle(_r, values_list):
    values = Array(*values_list)

    def fn():
        a = copy(values)
        random.shuffle(a)
        a.sort()
        b = copy(values)
        b.sort()
        return a == b

    assert run_compiled(fn)


@given(st.random_module())
def test_random(_r):
    def fn():
        value = random.random()
        assert_true(0.0 <= value < 1.0)
        return value

    result = run_compiled(fn)
    assert 0.0 <= result < 1.0


@given(st.random_module(), st.floats(min_value=-100.0, max_value=100.0), st.floats(min_value=0.0, max_value=100.0))
def test_uniform(_r, a, width):
    b = a + width

    def fn():
        value = random.uniform(a, b)
        assert_true(a <= value <= b)
        return value

    result = run_compiled(fn)
    assert a <= result <= b


MAX_ITERATIONS = 10000
MAX_RANGE_SIZE = 10


def collect_until_complete[T](
    generator: Callable[[], T], expected_values: set[T], max_iterations: int = MAX_ITERATIONS
) -> tuple[dict[T, int], int]:
    """Generate values until all expected values are generated or max iterations are reached."""
    counts: dict[T, int] = defaultdict(int)
    missing = expected_values.copy()
    iterations = 0

    while missing and iterations < max_iterations:
        value = generator()
        counts[value] += 1
        missing.discard(value)
        iterations += 1

    return counts, len(missing)


@given(st.random_module(), st.integers(min_value=1, max_value=MAX_RANGE_SIZE))
def test_impl_randrange_simple(_random_module, stop):
    expected = set(range(stop))
    counts, missing = collect_until_complete(lambda: srandom._randrange(stop), expected)

    assert missing == 0, f"Failed to generate values: {expected - set(counts.keys())}"


@given(
    st.random_module(),
    st.integers(min_value=-MAX_RANGE_SIZE, max_value=MAX_RANGE_SIZE),
    st.integers(min_value=1, max_value=MAX_RANGE_SIZE),
)
def test_impl_randrange_start_stop(_random_module, start, width):
    stop = start + width
    expected = set(range(start, stop))
    counts, missing = collect_until_complete(lambda: srandom._randrange(start, stop), expected)

    assert missing == 0, f"Failed to generate values: {expected - set(counts)}"
    assert all(start <= x < stop for x in counts)


@given(
    st.random_module(),
    st.integers(min_value=-MAX_RANGE_SIZE, max_value=MAX_RANGE_SIZE),
    st.integers(min_value=1, max_value=MAX_RANGE_SIZE),
    st.integers(min_value=1, max_value=5),
)
def test_impl_randrange_with_step(_random_module, start, width, step):
    stop = start + width
    expected = set(range(start, stop, step))
    if not expected:  # Empty range
        return

    counts, missing = collect_until_complete(lambda: srandom._randrange(start, stop, step), expected)

    assert missing == 0, f"Failed to generate values: {expected - set(counts)}"
    assert all((x - start) % step == 0 for x in counts)
    assert all(start <= x < stop for x in counts)


@given(
    st.random_module(),
    st.integers(min_value=-MAX_RANGE_SIZE, max_value=MAX_RANGE_SIZE),
    st.integers(min_value=0, max_value=MAX_RANGE_SIZE),
)
def test_impl_randint(_random_module, a, width):
    b = a + width
    expected = set(range(a, b + 1))
    counts, missing = collect_until_complete(lambda: srandom._randint(a, b), expected)

    assert missing == 0, f"Failed to generate values: {expected - set(counts.keys())}"
    assert all(a <= x <= b for x in counts)


@given(
    st.random_module(),
    st.lists(
        st.integers(min_value=-MAX_RANGE_SIZE, max_value=MAX_RANGE_SIZE),
        min_size=1,
        max_size=MAX_RANGE_SIZE,
        unique=True,
    ),
)
def test_impl_choice(_random_module, options):
    expected = set(options)
    counts, missing = collect_until_complete(lambda: srandom._choice(options), expected)

    assert missing == 0, f"Failed to generate values: {expected - set(counts.keys())}"


@given(
    st.random_module(),
    st.lists(st.integers(min_value=-MAX_RANGE_SIZE, max_value=MAX_RANGE_SIZE), min_size=1, max_size=5, unique=True),
)
def test_impl_shuffle(_random_module, original):
    # For small lists, we can test all permutations
    def get_permutations(lst: list[int]) -> set[tuple]:
        if len(lst) <= 1:
            return {tuple(lst)}
        result = set()
        for i in range(len(lst)):
            result.update((lst[i], *perm) for perm in get_permutations(lst[:i] + lst[i + 1 :]))
        return result

    expected_permutations = get_permutations(original)

    def generate_shuffle():
        test_impl_list = original.copy()
        srandom._shuffle(test_impl_list)
        return tuple(test_impl_list)

    counts, missing = collect_until_complete(generate_shuffle, expected_permutations)

    assert missing == 0, f"Failed to generate permutations: {expected_permutations - set(counts.keys())}"


@given(
    st.random_module(),
)
def test_impl_random_distribution(_random_module):
    # Test if random() generates values across all buckets
    num_buckets = 10
    bucket_size = 1.0 / num_buckets
    expected = set(range(num_buckets))

    def get_bucket():
        value = srandom._random()
        assert 0 <= value < 1, f"Value {value} outside [0,1)"
        return int(value / bucket_size)

    counts, missing = collect_until_complete(get_bucket, expected)

    assert missing == 0, f"Failed to generate values in buckets: {expected - set(counts.keys())}"


@given(
    st.random_module(),
    st.floats(min_value=-MAX_RANGE_SIZE, max_value=MAX_RANGE_SIZE),
    st.floats(min_value=0.1, max_value=MAX_RANGE_SIZE),
)
def test_impl_uniform_distribution(_random_module, a, width):
    num_buckets = 10
    b = a + width
    bucket_size = width / num_buckets
    expected = set(range(num_buckets))

    def get_bucket():
        value = srandom._uniform(a, b)
        assert a <= value <= b, f"Value {value} outside [{a},{b}]"
        bucket = int((value - a) / bucket_size)
        return min(bucket, num_buckets - 1)  # Clamp to last bucket if value is b

    counts, missing = collect_until_complete(get_bucket, expected)

    assert missing == 0, f"Failed to generate values in buckets: {expected - set(counts.keys())}"



from sonolus.script.internal.context import ctx
from sonolus.script.internal.generic import validate_concrete_type
from sonolus.script.internal.impl import validate_value
from sonolus.script.internal.meta_fn import meta_fn


@meta_fn
def alloc[T](type_: type[T]) -> T:  # type: ignore
    """Return an uninitialized instance of the given type.

    Use this carefully as reading from uninitialized memory can lead to unexpected behavior.
    """
    type_ = validate_concrete_type(type_)
    if ctx():
        return type_._alloc_()  # type: ignore
    else:
        return type_._alloc_()._as_py_()  # type: ignore


@meta_fn
def zeros[T](type_: type[T]) -> T:  # type: ignore
    """Make a new instance of the given type initialized with zeros.

    Generally works the same as the unary `+` operator on record and array types.
    """
    return validate_concrete_type(type_)._zero_()  # type: ignore


@meta_fn
def copy[T](value: T) -> T:
    """Make a deep copy of the given value.

    Generally works the same as the unary `+` operator on records and arrays.
    """
    value = validate_value(value)  # type: ignore
    if ctx():
        return value._copy_()  # type: ignore
    else:
        return value._copy_()._as_py_()  # type: ignore


def swap[T](a: T, b: T):
    """Swap the values of the two provided mutable values."""
    temp = copy(a)
    a @= b  # type: ignore
    b @= temp  # type: ignore


@meta_fn
def sizeof(type_: type, /) -> int:
    """Return the size of the given type."""
    return validate_concrete_type(type_)._size_()



from enum import Enum
from functools import cached_property

from sonolus.backend.blocks import Block, PlayBlock, PreviewBlock, TutorialBlock, WatchBlock


class Mode(Enum):
    blocks: type[Block]

    PLAY = (PlayBlock,)
    WATCH = (WatchBlock,)
    PREVIEW = (PreviewBlock,)
    TUTORIAL = (TutorialBlock,)

    def __init__(self, blocks: type[Block]):
        self.blocks = blocks

    @cached_property
    def callbacks(self) -> frozenset[str]:
        cbs = set()
        for block in self.blocks:
            cbs.update(block.readable)
            cbs.update(block.writable)
        return frozenset(cbs)



from sonolus.script.array import Dim
from sonolus.script.containers import ArraySet
from sonolus.script.stream import Stream, StreamGroup, streams


@streams
class Streams:
    # Records the set of lanes at each time when the empty tap lane effect was played.
    effect_lanes: Stream[ArraySet[float, Dim[16]]]

    # Records whether a hold is active at a given time.
    # Keyed by the hold head's index.
    hold_activity: StreamGroup[bool, Dim[99999]]



from __future__ import annotations

from collections.abc import Iterable, Sequence
from contextlib import contextmanager
from dataclasses import dataclass
from enum import Enum
from threading import Lock
from typing import TYPE_CHECKING, Any, Literal, Self

from sonolus.backend.blocks import BlockData, PlayBlock
from sonolus.backend.ir import IRConst, IRExpr, IRStmt
from sonolus.backend.mode import Mode
from sonolus.backend.optimize.flow import BasicBlock, FlowEdge
from sonolus.backend.place import (
    PREALLOCATED_TEMP_PLACE_COUNT,
    Block,
    BlockPlace,
    TempBlock,
    preallocated_temp_block_places,
)
from sonolus.script.internal.value import Value

if TYPE_CHECKING:
    from sonolus.script.globals import _GlobalInfo, _GlobalPlaceholder
    from sonolus.script.project import BuildConfig

_compiler_internal_ = True

_context: Context | None = None


@dataclass(frozen=True)
class DebugConfig:
    unchecked_reads: bool
    unchecked_writes: bool


_full_debug_config = DebugConfig(
    unchecked_reads=True,
    unchecked_writes=True,
)

_disabled_debug_config = DebugConfig(
    unchecked_reads=False,
    unchecked_writes=False,
)

_debug_config: DebugConfig = _disabled_debug_config


class RuntimeChecks(Enum):
    """Runtime error checking modes."""

    NONE = "none"
    """No runtime checks."""

    TERMINATE = "terminate"
    """Terminate on errors."""

    NOTIFY_AND_TERMINATE = "notify_and_terminate"
    """Log, debug pause, and terminate on errors."""


@dataclass
class FunctionVisitStatistics:
    total_time: int = 0
    own_time: int = 0
    call_count: int = 0


class ProjectContextState:
    rom: ReadOnlyMemory
    const_mappings: dict[Any, int]
    debug_str_mappings: dict[str, int]
    lock: Lock
    runtime_checks: RuntimeChecks
    visit_stats: dict[str, FunctionVisitStatistics]

    def __init__(
        self,
        rom: ReadOnlyMemory | None = None,
        const_mappings: dict[Any, int] | None = None,
        debug_str_mappings: dict[str, int] | None = None,
        runtime_checks: RuntimeChecks = RuntimeChecks.NONE,
    ):
        self.rom = ReadOnlyMemory() if rom is None else rom
        self.const_mappings = {} if const_mappings is None else const_mappings
        self.debug_str_mappings = {} if debug_str_mappings is None else debug_str_mappings
        self.lock = Lock()
        self.runtime_checks = runtime_checks
        self.visit_stats = {}

    @classmethod
    def from_build_config(
        cls,
        config: BuildConfig,
        rom: ReadOnlyMemory | None = None,
        const_mappings: dict[Any, int] | None = None,
        debug_str_mappings: dict[str, int] | None = None,
    ) -> Self:
        return cls(
            rom=rom,
            const_mappings=const_mappings,
            debug_str_mappings=debug_str_mappings,
            runtime_checks=config.runtime_checks,
        )


class ModeContextState:
    archetypes: dict[type, int]
    compile_time_only_archetypes: set[type]
    archetypes_by_name: dict[str, type]
    keys_by_archetype_id: Sequence[int]
    is_scored_by_archetype_id: Sequence[bool]
    archetype_mro_id_array_rom_indexes: Sequence[int] | None = None
    environment_mappings: dict[_GlobalInfo, int]
    environment_offsets: dict[Block, int]
    mode: Mode
    lock: Lock

    def __init__(self, mode: Mode, archetypes: list[type] | None = None):
        from sonolus.script.array import Array

        archetypes = [*archetypes] if archetypes is not None else []
        seen_archetypes = {*archetypes}
        compile_time_only_archetypes = set()
        for type_ in [*archetypes]:
            for entry in type_.mro():
                if getattr(entry, "_is_concrete_archetype_", False) and entry not in seen_archetypes:
                    archetypes.append(entry)
                    seen_archetypes.add(entry)
                    compile_time_only_archetypes.add(entry)
        self.archetypes = {type_: idx for idx, type_ in enumerate(archetypes)}
        self.compile_time_only_archetypes = compile_time_only_archetypes
        self.archetypes_by_name = {type_.name: type_ for type_, _ in self.archetypes.items()}  # type: ignore
        self.keys_by_archetype_id = (
            Array(*((getattr(a, "_key_", -1)) for a in archetypes)) if archetypes else Array[int, Literal[0]]()
        )
        self.is_scored_by_archetype_id = (
            Array(*((getattr(a, "_is_scored_", False)) for a in archetypes))
            if archetypes
            else Array[bool, Literal[0]]()
        )
        self.environment_mappings = {}
        self.environment_offsets = {}
        self.mode = mode
        self.lock = Lock()

    def _init_archetype_mro_info(self, rom: ReadOnlyMemory):
        from sonolus.script.array import Array
        from sonolus.script.num import Num

        with self.lock:
            if self.archetype_mro_id_array_rom_indexes is not None:
                return
            archetype_mro_id_values = []
            archetype_mro_id_offsets = []
            for type_ in self.archetypes:
                mro_ids = [self.archetypes[entry] for entry in type_.mro() if entry in self.archetypes]
                archetype_mro_id_offsets.append(len(archetype_mro_id_values))
                archetype_mro_id_values.append(len(mro_ids))
                archetype_mro_id_values.extend(mro_ids)
            archetype_mro_id_array_place = rom[tuple(archetype_mro_id_values)]

            archetype_mro_id_rom_indexes = Array[int, len(archetype_mro_id_offsets)]._with_value(
                [Num._accept_(offset + archetype_mro_id_array_place.index) for offset in archetype_mro_id_offsets]
            )
            self.archetype_mro_id_array_rom_indexes = archetype_mro_id_rom_indexes


class CallbackContextState:
    callback: str
    used_names: dict[str, int]
    debug_stack: list[str]
    no_eval: bool
    visitor_own_time: int

    def __init__(self, callback: str, no_eval: bool = False):
        self.callback = callback
        self.used_names = {}
        self.debug_stack = []
        self.no_eval = no_eval
        self.visitor_own_time = 0


class Context:
    project_state: ProjectContextState
    mode_state: ModeContextState
    callback_state: CallbackContextState
    statements: list[IRStmt]
    test: IRExpr
    outgoing: dict[float | None, Context]
    scope: Scope
    loop_variables: dict[str, Value]
    live: bool

    def __init__(
        self,
        project_state: ProjectContextState,
        mode_state: ModeContextState,
        callback_state: CallbackContextState,
        scope: Scope | None = None,
        live: bool = True,
        foldable_constants: dict[TempBlock, list[float | None]] | None = None,
    ):
        self.project_state = project_state
        self.mode_state = mode_state
        self.callback_state = callback_state
        self.statements = []
        self.test = IRConst(0)
        self.outgoing = {}
        self.scope = scope if scope is not None else Scope()
        self.loop_variables = {}
        self.live = live
        self.foldable_constants = foldable_constants if foldable_constants is not None else {}

    @property
    def rom(self) -> ReadOnlyMemory:
        return self.project_state.rom

    @property
    def blocks(self) -> type[Block]:
        return self.mode_state.mode.blocks

    @property
    def callback(self) -> str:
        return self.callback_state.callback

    @property
    def used_names(self) -> dict[str, int]:
        return self.callback_state.used_names

    @property
    def no_eval(self) -> bool:
        return self.callback_state.no_eval

    def check_readable(self, place: BlockPlace):
        if debug_config().unchecked_reads:
            return
        if not self.callback:
            return
        if isinstance(place.block, BlockData) and not self.is_readable(place):
            raise RuntimeError(f"Block {place.block} is not readable in {self.callback}")

    def check_writable(self, place: BlockPlace):
        if debug_config().unchecked_writes:
            return
        if not self.callback:
            return
        if isinstance(place.block, BlockData) and not self.is_writable(place):
            raise RuntimeError(f"Block {place.block} is not writable in {self.callback}")

    def is_readable(self, place: BlockPlace) -> bool:
        if debug_config().unchecked_reads:
            return True
        return self.callback and self.callback in self.blocks(place.block).readable

    def is_writable(self, place: BlockPlace) -> bool:
        if debug_config().unchecked_writes:
            return True
        return self.callback and self.callback in self.blocks(place.block).writable

    def add_statement(self, statement: IRStmt):
        if not self.live:
            return
        self.statements.append(statement)

    def add_statements(self, *statements: IRStmt):
        for statement in statements:
            self.add_statement(statement)

    def alloc(self, name: str | None = None, size: int = 1) -> BlockPlace:
        if size == 0:
            return BlockPlace(TempBlock(name or "e", 0), 0)
        name = name or ("v" if size == 1 else "a")
        num = self._get_alloc_number(name)
        if name == "v" and num < PREALLOCATED_TEMP_PLACE_COUNT:
            return preallocated_temp_block_places[num]
        return BlockPlace(TempBlock(f"{name}{num}", size), 0)

    def _get_alloc_number(self, name: str) -> int:
        if name not in self.used_names:
            self.used_names[name] = 0
        self.used_names[name] += 1
        return self.used_names[name]

    def save_alloc_state(self) -> dict[str, int]:
        return self.used_names.copy()

    def restore_alloc_state(self, state: dict[str, int]):
        self.callback_state.used_names = state.copy()

    def copy_with_scope(self, scope: Scope) -> Context:
        return Context(
            project_state=self.project_state,
            mode_state=self.mode_state,
            callback_state=self.callback_state,
            scope=scope,
            live=self.live,
            foldable_constants={**self.foldable_constants},  # We want this to persist even if the scope changes
        )

    def branch(self, condition: float | None):
        assert condition not in self.outgoing
        result = self.copy_with_scope(self.scope.copy())
        self.outgoing[condition] = result
        return result

    def branch_with_scope(self, condition: float | None, scope: Scope):
        assert condition not in self.outgoing
        result = self.copy_with_scope(scope)
        self.outgoing[condition] = result
        return result

    def new_disconnected(self):
        return self.copy_with_scope(self.scope.copy())

    def new_empty_disconnected(self):
        return self.copy_with_scope(Scope())

    def into_dead(self):
        """Create a new context for code that is unreachable, like after a return statement."""
        result = self.copy_with_scope(self.scope.copy())
        result.live = False
        return result

    def prepare_loop_header(self, to_merge: set[str]) -> Context:
        # to_merge is the set of bindings set anywhere in the loop
        # we need to invalidate them in the header if they're reference types
        # or merge them if they're value types
        # structure is self -> intermediate -> header -> body (continue -> header) | exit
        assert len(self.outgoing) == 0
        header = self.branch(None)
        for name in sorted(to_merge):
            binding = self.scope.get_binding(name)
            if not isinstance(binding, ValueBinding):
                continue
            value = binding.value
            type_ = type(value)
            if type_._is_value_type_():
                target_value = type_._from_place_(header.alloc(size=type_._size_()))
                with using_ctx(self):
                    target_value._set_(value)
                header.scope.set_value(name, target_value)
                header.loop_variables[name] = target_value
            else:
                header.scope.set_value(name, value)
                header.loop_variables[name] = value
        return header

    def branch_to_loop_header(self, header: Self):
        if not self.live:
            return
        assert len(self.outgoing) == 0
        self.outgoing[None] = header
        values = {}
        # First do a pass through and get every value
        for name, target_value in header.loop_variables.items():
            with using_ctx(self):
                if type(target_value)._is_value_type_():
                    value = self.scope.get_value(name)
                    # We make this call to _get_readonly_() to ensure that we're reading the value at this
                    # point in time specifically, since _get_readonly_ will make a copy if the value is
                    # e.g. a Num backed by a TempBlock which could be mutated.
                    values[name] = value._get_readonly_()
        # Then actually set them
        for name, target_value in header.loop_variables.items():
            with using_ctx(self):
                if type(target_value)._is_value_type_():
                    value = values[name]
                    value = type(target_value)._accept_(value)
                    target_value._set_(value)
                else:
                    value = self.scope.get_value(name)
                    if target_value is not value:
                        raise RuntimeError(
                            f"Variable '{name}' may have conflicting definitions between loop iterations"
                        )

    def map_constant(self, value: Any) -> int:
        with self.project_state.lock:
            const_mappings = self.project_state.const_mappings
            if value not in const_mappings:
                const_mappings[value] = len(const_mappings)
            return const_mappings[value]

    def map_debug_message(self, message: str) -> int:
        with self.project_state.lock:
            message_with_trace = "\n".join([*self.callback_state.debug_stack, message])
            debug_str_mappings = self.project_state.debug_str_mappings
            if message_with_trace not in debug_str_mappings:
                debug_str_mappings[message_with_trace] = len(debug_str_mappings) + 1
            return debug_str_mappings[message_with_trace]

    def get_global_base(self, value: _GlobalInfo | _GlobalPlaceholder) -> BlockPlace:
        with self.mode_state.lock:
            block = value.blocks.get(self.mode_state.mode)
            if block is None:
                raise RuntimeError(f"Global {value} is not available in '{self.mode_state.mode.name}' mode")
            if value not in self.mode_state.environment_mappings:
                if value.offset is None:
                    offset = self.mode_state.environment_offsets.get(block, 0)
                    self.mode_state.environment_mappings[value] = offset
                    self.mode_state.environment_offsets[block] = offset + value.size
                else:
                    self.mode_state.environment_mappings[value] = value.offset
            return BlockPlace(block, self.mode_state.environment_mappings[value])

    @classmethod
    def meet(cls, contexts: list[Context]) -> Context:
        if not contexts:
            raise RuntimeError("Cannot meet empty list of contexts")
        if not any(context.live for context in contexts):
            return contexts[0].into_dead()
        contexts = [context for context in contexts if context.live]
        assert not any(context.outgoing for context in contexts)
        assert all(len(context.outgoing) == 0 for context in contexts)
        target = contexts[0].copy_with_scope(Scope())
        Scope.apply_merge(target, contexts)
        for context in contexts:
            context.outgoing[None] = target
        return target

    def register_archetype(self, type_: type) -> int:
        with self.mode_state.lock:
            if type_ not in self.mode_state.archetypes:
                self.mode_state.archetypes[type_] = len(self.mode_state.archetypes)
            return self.mode_state.archetypes[type_]

    def get_archetype_mro_id_array(self, archetype_id: int) -> Sequence[int]:
        from sonolus.script.containers import ArrayPointer
        from sonolus.script.num import Num
        from sonolus.script.pointer import _deref

        self.mode_state._init_archetype_mro_info(self.rom)
        rom_index = self.mode_state.archetype_mro_id_array_rom_indexes[archetype_id]
        return ArrayPointer[int](_deref(self.blocks.EngineRom, rom_index, Num), self.blocks.EngineRom, rom_index + 1)


def ctx() -> Context | Any:  # Using Any to silence type checker warnings if it's None
    return _context


def set_ctx(value: Context | None):
    global _context  # noqa: PLW0603
    old_value = _context
    _context = value
    return old_value


@contextmanager
def using_ctx(value: Context | None):
    global _context  # noqa: PLW0603
    old_value = _context
    _context = value
    try:
        yield
    finally:
        _context = old_value


class ReadOnlyMemory:
    values: list[float]
    indexes: dict[tuple[float, ...], int]
    _lock: Lock

    def __init__(self):
        self.values = [
            float("nan"),
            float("inf"),
            float("-inf"),
        ]
        self.indexes = {}
        self._lock = Lock()

    def __getitem__(self, item: tuple[float, ...]) -> BlockPlace:
        with self._lock:
            if item not in self.indexes:
                index = len(self.values)
                self.indexes[item] = index
                self.values.extend(item)
            else:
                index = self.indexes[item]
            return BlockPlace(self.block, index)

    @property
    def block(self) -> Block:
        if ctx():
            return ctx().blocks.EngineRom
        else:
            return PlayBlock.EngineRom

    def get_value(self, index: int) -> float:
        with self._lock:
            if index < 0 or index >= len(self.values):
                raise IndexError(f"Index {index} out of bounds for ReadOnlyMemory")
            return self.values[index]


@contextmanager
def enable_debug(config: DebugConfig | None = None):
    global _debug_config  # noqa: PLW0603
    if config is None:
        config = _full_debug_config
    old_config = _debug_config
    _debug_config = config
    try:
        yield
    finally:
        _debug_config = old_config


def debug_config() -> DebugConfig:
    return _debug_config


@dataclass
class ValueBinding:
    value: Value


@dataclass
class ConflictBinding:
    pass


@dataclass
class EmptyBinding:
    pass


Binding = ValueBinding | ConflictBinding | EmptyBinding

_EMPTY_BINDING = EmptyBinding()


class Scope:
    bindings: dict[str, Binding]

    def __init__(self, bindings: dict[str, Binding] | None = None):
        self.bindings = bindings or {}

    def get_binding(self, name: str) -> Binding:
        return self.bindings.get(name, _EMPTY_BINDING)

    def set_binding(self, name: str, binding: Binding):
        self.bindings[name] = binding

    def get_value(self, name: str) -> Value | Any:
        binding = self.get_binding(name)
        match binding:
            case ValueBinding(value):
                # we don't need to call _get_() here because _set_() is never called where it could be a problem
                return value
            case ConflictBinding():
                raise RuntimeError(
                    f"Binding '{name}' has multiple conflicting definitions or may not be guaranteed to be defined"
                )
            case EmptyBinding():
                raise RuntimeError(f"Binding '{name}' is not defined")

    def set_value(self, name: str, value: Value):
        from sonolus.script.internal.impl import validate_value

        self.bindings[name] = ValueBinding(validate_value(value))

    def delete_binding(self, name: str):
        del self.bindings[name]

    def copy(self) -> Scope:
        return Scope(self.bindings.copy())

    @classmethod
    def apply_merge(cls, target: Context, incoming: list[Context]):
        if not incoming:
            return
        assert all(len(inc.outgoing) == 0 for inc in incoming)
        sources = [context.scope for context in incoming]
        keys = unique(key for source in sources for key in source.bindings)
        for key in keys:
            bindings = [source.get_binding(key) for source in sources]
            if not all(isinstance(binding, ValueBinding) for binding in bindings):
                target.scope.set_binding(key, ConflictBinding())
                continue
            values = [binding.value for binding in bindings]
            if len({id(value) for value in values}) == 1:
                target.scope.set_binding(key, ValueBinding(values[0]))
                continue
            types = {type(value) for value in values}
            if len(types) > 1:
                target.scope.set_binding(key, ConflictBinding())
                continue
            common_type: type[Value] = types.pop()
            with using_ctx(target):
                target_value = common_type._get_merge_target_(values)
            if target_value is not NotImplemented:
                for inc in incoming:
                    with using_ctx(inc):
                        target_value._set_(inc.scope.get_value(key))
                target.scope.set_value(key, target_value)
                continue
            else:
                target.scope.set_binding(key, ConflictBinding())
                continue


def iter_contexts(context: Context):
    seen = set()
    queue = [context]
    while queue:
        current = queue.pop()
        if current in seen:
            continue
        seen.add(current)
        yield current
        queue.extend(current.outgoing.values())


def context_to_cfg(context: Context) -> BasicBlock:
    blocks = {context: BasicBlock(statements=context.statements, test=context.test)}
    for current in iter_contexts(context):
        for condition, target in current.outgoing.items():
            if target not in blocks:
                blocks[target] = BasicBlock(statements=target.statements, test=target.test)
            edge = FlowEdge(src=blocks[current], dst=blocks[target], cond=condition)
            blocks[current].outgoing.add(edge)
            blocks[target].incoming.add(edge)
    result = blocks[context]
    for current in tuple(iter_contexts(context)):
        # Break cycles so memory can be cleaned without gc
        del current.outgoing
    return result


def unique[T](iterable: Iterable[T]) -> list[T]:
    result = []
    seen = set()
    for item in iterable:
        if item not in seen:
            seen.add(item)
            result.append(item)
    return result



from sonolus.script.runtime import HorizontalAlign, runtime_ui, screen
from sonolus.script.ui import UiConfig, UiJudgmentErrorPlacement, UiJudgmentErrorStyle
from sonolus.script.vec import Vec2

ui_config = UiConfig(
    judgment_error_style=UiJudgmentErrorStyle.LATE,
    judgment_error_placement=UiJudgmentErrorPlacement.TOP,
    judgment_error_min=20.0,
)


def init_ui():
    ui = runtime_ui()
    ui.menu.update(
        anchor=screen().tr + Vec2(-0.05, -0.05),
        pivot=Vec2(1, 1),
        dimensions=Vec2(0.15, 0.15) * ui.menu_config.scale,
        rotation=0,
        alpha=ui.menu_config.alpha,
        horizontal_align=HorizontalAlign.CENTER,
        background=True,
    )
    ui.judgment.update(
        anchor=Vec2(0, -0.25),
        pivot=Vec2(0.5, 0),
        dimensions=Vec2(0, 0.15) * ui.judgment_config.scale,
        rotation=0,
        alpha=ui.judgment_config.alpha,
        horizontal_align=HorizontalAlign.CENTER,
        background=False,
    )
    ui.combo_value.update(
        anchor=Vec2(screen().r * 0.7, 0),
        pivot=Vec2(0.5, 0),
        dimensions=Vec2(0, 0.2) * ui.combo_config.scale,
        rotation=0,
        alpha=ui.combo_config.alpha,
        horizontal_align=HorizontalAlign.CENTER,
        background=False,
    )
    ui.combo_text.update(
        anchor=Vec2(screen().r * 0.7, 0),
        pivot=Vec2(0.5, 1),
        dimensions=Vec2(0, 0.12) * ui.combo_config.scale,
        rotation=0,
        alpha=ui.combo_config.alpha,
        horizontal_align=HorizontalAlign.CENTER,
        background=False,
    )
    ui.primary_metric_bar.update(
        anchor=screen().tl + Vec2(0.05, -0.05),
        pivot=Vec2(0, 1),
        dimensions=Vec2(0.75, 0.15) * ui.primary_metric_config.scale,
        rotation=0,
        alpha=ui.primary_metric_config.alpha,
        horizontal_align=HorizontalAlign.LEFT,
        background=True,
    )
    ui.primary_metric_value.update(
        anchor=screen().tl + Vec2(0.05, -0.05) + Vec2(0.715, -0.035) * ui.primary_metric_config.scale,
        pivot=Vec2(0, 1),
        dimensions=Vec2(0, 0.08) * ui.primary_metric_config.scale,
        rotation=0,
        alpha=ui.primary_metric_config.alpha,
        horizontal_align=HorizontalAlign.RIGHT,
        background=False,
    )
    ui.secondary_metric_bar.update(
        anchor=screen().tr - Vec2(0.05, 0.05) - Vec2(0.05, 0) - Vec2(0.15, 0) * ui.menu_config.scale,
        pivot=Vec2(1, 1),
        dimensions=Vec2(0.75, 0.15) * ui.secondary_metric_config.scale,
        rotation=0,
        alpha=ui.secondary_metric_config.alpha,
        horizontal_align=HorizontalAlign.LEFT,
        background=True,
    )
    ui.secondary_metric_value.update(
        anchor=screen().tr
        - Vec2(0.05, 0.05)
        - Vec2(0.05, 0)
        - Vec2(0.15, 0) * ui.menu_config.scale
        - Vec2(0.035, 0.035) * ui.primary_metric_config.scale,
        pivot=Vec2(1, 1),
        dimensions=Vec2(0, 0.08) * ui.secondary_metric_config.scale,
        rotation=0,
        alpha=ui.secondary_metric_config.alpha,
        horizontal_align=HorizontalAlign.RIGHT,
        background=False,
    )
    ui.progress.update(
        anchor=screen().bl + Vec2(0.05, 0.05),
        pivot=Vec2(0, 0),
        dimensions=Vec2(screen().w - 0.1, 0.15 * ui.progress_config.scale),
        rotation=0,
        alpha=ui.progress_config.alpha,
        horizontal_align=HorizontalAlign.CENTER,
        background=True,
    )
    ui.previous.update(
        anchor=Vec2(screen().l + 0.05, 0),
        pivot=Vec2(0, 0.5),
        dimensions=Vec2(0.15, 0.15) * ui.navigation_config.scale,
        rotation=0,
        alpha=ui.navigation_config.alpha,
        background=True,
    )
    ui.next.update(
        anchor=Vec2(screen().r - 0.05, 0),
        pivot=Vec2(1, 0.5),
        dimensions=Vec2(0.15, 0.15) * ui.navigation_config.scale,
        rotation=0,
        alpha=ui.navigation_config.alpha,
        background=True,
    )
    ui.instruction.update(
        anchor=Vec2(0, 0.2),
        pivot=Vec2(0.5, 0.5),
        dimensions=Vec2(1.2, 0.15) * ui.instruction_config.scale,
        rotation=0,
        alpha=ui.instruction_config.alpha,
        background=True,
    )
